# Copyright (c) 2018-2020 Mika Tuupola
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of  this software and associated documentation files (the "Software"), to
# deal in  the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copied of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Branca

Authenticated and encrypted API tokens using modern crypto.
"""

import base62
import calendar
import ctypes
import binascii
import struct
from datetime import datetime
from xchacha20poly1305 import generate_nonce
from xchacha20poly1305 import crypto_aead_xchacha20poly1305_ietf_encrypt
from xchacha20poly1305 import crypto_aead_xchacha20poly1305_ietf_decrypt
from xchacha20poly1305 import CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES
from xchacha20poly1305 import CRYPTO_AEAD_XHCACHA20POLY1305_IETF_KEYBYTES

class Branca:
    VERSION = 0xBA

    def __init__(self, key):
        if isinstance(key, bytes):
            self._key = key
        else:
            self._key = key.encode()

        if len(key) is not CRYPTO_AEAD_XHCACHA20POLY1305_IETF_KEYBYTES:
            raise ValueError(
                "Secrect key should be {} bytes long".format(
                    CRYPTO_AEAD_XHCACHA20POLY1305_IETF_KEYBYTES
                )
            )

        self._nonce = None # Used only for unit testing!

    def encode(self, payload, timestamp=None):

        if not isinstance(payload, bytes):
            payload = payload.encode()

        if timestamp is None:
            timestamp = calendar.timegm(datetime.utcnow().timetuple())

        version = struct.pack("B", self.VERSION)
        time = struct.pack(">L", timestamp)

        if self._nonce is None:
            nonce = generate_nonce()
        else:
            nonce = self._nonce

        header = version + time + nonce
        ciphertext = crypto_aead_xchacha20poly1305_ietf_encrypt(payload, header, nonce, self._key)

        return base62.encodebytes(header + ciphertext)

    def decode(self, token, ttl=None):
        token = base62.decodebytes(token)
        header = token[0:CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES + 5]
        nonce = header[5:CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES + 5]
        ciphertext = token[CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES + 5:]

        version, time = struct.unpack(">BL", bytes(header[0:5]))

        # Implementation should accept only current version.
        if version is not self.VERSION:
            raise RuntimeError("Invalid token version")

        payload = crypto_aead_xchacha20poly1305_ietf_decrypt(ciphertext, header, nonce, self._key)

        if ttl is not None:
            future = time + ttl
            timestamp = calendar.timegm(datetime.utcnow().timetuple())
            if future < timestamp:
                raise RuntimeError("Token is expired")

        return payload

    def timestamp(self, token):
        token = base62.decodebytes(token)
        version, time = struct.unpack(">BL", bytes(token[0:5]))

        return time


# Wrapper for libsodium IETF XChaCha20-Poly1305 AEAD
#
# Copyright (c) 2013-2018, Marsiske Stefan.
# Copyright (c) 2018-2020 Mika Tuupola.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
IETF XChaCha20-Poly1305 AEAD

Wrapper for libsodium IETF XChaCha20-Poly1305 AEAD functions.
"""

import ctypes
import ctypes.util

library_path = ctypes.util.find_library("sodium") or ctypes.util.find_library("libsodium")
sodium = ctypes.cdll.LoadLibrary(library_path)

if not sodium._name:
    raise RuntimeError("Unable to locate libsodium")

CRYPTO_AEAD_XHCACHA20POLY1305_IETF_KEYBYTES = sodium.crypto_aead_xchacha20poly1305_ietf_keybytes()
CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES = sodium.crypto_aead_xchacha20poly1305_ietf_npubbytes()
CRYPTO_AEAD_XHCACHA20POLY1305_IETF_ABYTES = sodium.crypto_aead_xchacha20poly1305_ietf_abytes()

# crypto_aead_xchacha20poly1305_ietf_encrypt(ciphertext, &ciphertext_len,
#                                           MESSAGE, MESSAGE_LEN,
#                                           ADDITIONAL_DATA, ADDITIONAL_DATA_LEN,
#                                           NULL, nonce, key);

def crypto_aead_xchacha20poly1305_ietf_encrypt(message, ad, nonce, key):
    if len(nonce) is not CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES:
        raise ValueError("Invalid nonce")

    if len(key) is not CRYPTO_AEAD_XHCACHA20POLY1305_IETF_KEYBYTES:
        raise ValueError("Invalid key")

    message_len = ctypes.c_ulonglong(len(message))

    if ad is None:
        ad_len = ctypes.c_ulonglong(0)
    else:
        ad_len = ctypes.c_ulonglong(len(ad))

    ciphertext = ctypes.create_string_buffer(
        message_len.value + CRYPTO_AEAD_XHCACHA20POLY1305_IETF_ABYTES
    )
    ciphertext_len = ctypes.c_ulonglong(0)

    retval = sodium.crypto_aead_xchacha20poly1305_ietf_encrypt(
        ciphertext, ctypes.byref(ciphertext_len),
        message, message_len,
        ad, ad_len,
        None, nonce, key
    )

    if retval != 0:
        raise RuntimeError("Encrypting token failed")

    return ciphertext.raw

# if (crypto_aead_xchacha20poly1305_ietf_decrypt(decrypted, &decrypted_len,
#                                               NULL,
#                                               ciphertext, ciphertext_len,
#                                               ADDITIONAL_DATA,
#                                               ADDITIONAL_DATA_LEN,
#                                               nonce, key) != 0) {

def crypto_aead_xchacha20poly1305_ietf_decrypt(ciphertext, ad, nonce, key):
    if len(nonce) != CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES:
        raise ValueError("Invalid nonce")

    if len(key) != CRYPTO_AEAD_XHCACHA20POLY1305_IETF_KEYBYTES:
        raise ValueError("Invalid key")

    decrypted = ctypes.create_string_buffer(
        len(ciphertext) - CRYPTO_AEAD_XHCACHA20POLY1305_IETF_ABYTES
    )
    decrypted_len = ctypes.c_ulonglong(0)
    ciphertext_len = ctypes.c_ulonglong(len(ciphertext))

    if ad is None:
        ad_len = ctypes.c_ulonglong(0)
    else:
        ad_len = ctypes.c_ulonglong(len(ad))

    retval = sodium.crypto_aead_xchacha20poly1305_ietf_decrypt(
        decrypted, ctypes.byref(decrypted_len),
        None,
        ciphertext, ciphertext_len,
        ad, ad_len,
        nonce, key
    )

    if retval != 0:
        raise RuntimeError("Decrypting token failed")

    return decrypted.raw

def generate_nonce():
    buffer = ctypes.create_string_buffer(CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES)
    sodium.randombytes(buffer, ctypes.c_ulonglong(CRYPTO_AEAD_XHCACHA20POLY1305_IETF_NPUBBYTES))
    return buffer.raw


from setuptools import setup
from codecs import open
from os import path

cwd = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(cwd, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pybranca",
    py_modules=["branca", "xchacha20poly1305"],
    version="0.4.0",
    description="Authenticated and encrypted API tokens using modern crypto",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="api, token, jwt, xchacha20, poly1305",
    url="https://github.com/tuupola/pybranca",
    author="Mika Tuupola",
    author_email="tuupola@appelsiini.net",
    maintainer="Mika Tuupola",
    maintainer_email="tuupola@appelsiini.net",
    license="MIT",
    install_requires=[
        "pybase62>=0.3"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Topic :: Security",
        "License :: OSI Approved :: MIT License",
    ],
)
